import os
import pandas as pd
import numpy as np
from pandas.errors import EmptyDataError
import itertools

class Mlst:
	"""
	Use multi-locus sequence typing (MLST) to assign genome to sequence type (ST)

	at2st
	purpose: selects most likely ST from a list of allelic types (ATs)
	input:
		alleles = list of detected best-matching alleles, ordered alphabetically
		profiles = file containing profiles for each gene and the resulting ST
		perfect_matches = number of alleles which perfectly matched database (produced by parse_mlst in blast.py)
		final_results_directory = path to BRiskTyper final results directory
		prefix = genome prefix to use for output files
	output:
		predicted ST corresponding to best-matching alleles
	
	"""


	def __init__(self, alleles, profiles, perfect_matches, final_results_directory, prefix):

		self.alleles = alleles
		self.profiles = profiles
		self.perfect_matches = perfect_matches
		self.final_results_directory = final_results_directory
		self.prefix = prefix


	def at2st(self, alleles, profiles, perfect_matches, final_results_directory, prefix):

		if len(alleles) == 7:
			final = []	
			profiles = pd.read_csv(profiles, sep = "\t", header = 0)	
			alleles = list(itertools.product(*alleles))
			for allele in alleles:
				st_rows = profiles[(profiles.glp == int(allele[0])) & (profiles.gmk == int(allele[1])) & (profiles.ilv == int(allele[2])) & (profiles.pta == int(allele[3])) & (profiles.pur == int(allele[4])) & (profiles.pyc == int(allele[5])) & (profiles.tpi == int(allele[6]))]	
				if st_rows.shape[0] > 0:
					if str(st_rows["clonal_complex"].iloc[0]) == "nan":
						cc = "No clonal complex"
					else:
						cc = str(st_rows["clonal_complex"].iloc[0])
					final.append(str(st_rows["ST"].iloc[0]) + "[" + cc + "](" + str(perfect_matches) + "/7)")
			if len(final) > 0:
				final = final
			else:
				final = ["Unknown(unknown ST)"]
		else:
			final = ["Unknown(missing alleles)"]

		return(";".join(final))



import csv
import io
import os
import gzip
import contextlib
import importlib.resources
import itertools
import subprocess
import tempfile
import warnings

import Bio.SeqIO
import pandas
import pyfastani
from pandas.errors import EmptyDataError

class Ani:
	"""
	Use ANI to assign genome to genospecies and/or subspecies

	run_fastani
	purpose: runs fastANI for species/subspecies assignment and selects match with highest ANI
	input:
		taxon = "species", "subspecies", "geneflow", or "cytotoxicity"; corresponds to directory of genomes to use
		fasta = query genome for fastANI
		final_results_directory = path to BRiskTyper final results directory
		prefix = genome prefix to use for output files
	output:
		species, subspecies, or strain producing highest ANI value

	"""

	def __init__(self, taxon, fasta, final_results_directory, prefix):
		self.taxon = taxon
		self.fasta = fasta
		self.final_results_directory = final_results_directory
		self.prefix = prefix

	def run_fastani(self, taxon, fasta, final_results_directory, prefix):
		# create output folder if it doesn't exist
		ani_results_dir = os.path.join(final_results_directory, taxon)
		os.makedirs(ani_results_dir, exist_ok=True)

		# create the FastANI sketch
		sketch = pyfastani.Sketch()

		# get the ANI database using `importlib.resources`: note that since
		# files may not be available on the local filesystem (e.g. they could
		# be zipped), we use `importlib.resources.path` to get a system path
		# for each of them than can be passed to `fastANI`.
		data_module = "brisktyper.seq_ani_db.{}".format(taxon)
		with importlib.resources.open_text(data_module, "{}.tsv".format(taxon)) as f:
			database = pandas.read_table(f, comment="#")
		# extract all the reference genomes
		for genome_id in database["id"]:
			with importlib.resources.open_binary(data_module, genome_id) as handle:
				reader = io.TextIOWrapper(gzip.GzipFile(fileobj=handle, mode="rb"))
				records = Bio.SeqIO.parse(reader, "fasta")
				sketch.add_draft(genome_id, (str(record.seq) for record in records))

		# index the references
		mapper = sketch.index()

		# query mapper with the input file
		with open(fasta) as handle:
			records = Bio.SeqIO.parse(handle, "fasta")
			sequences = [str(record.seq) for record in records]
			self.check_fragmentation(sequences, mapper.fragment_length)
			hits = mapper.query_draft(sequences)

		# make a table from the hits
		results = pandas.DataFrame(
			data=[
				[fasta, hit.name, hit.identity, hit.matches, hit.fragments]
				for hit in hits
			],
			columns=[
				"query",
				"hit",
				"ani",
				"matches",
				"fragments",
			],
		)

		# write raw FastANI results
		result_file = os.path.join(ani_results_dir, "{}_{}_fastani.txt".format(prefix, taxon))
		results.to_csv(result_file, index=False, header=False, sep="\t")

		# left join with database to get associated metadata for each hit
		results = pandas.merge( results, database, how="left", left_on="hit", right_on="id")

		if not results.empty:
			# sort values by decreasing ANI and get best hit
			results.sort_values("ani", ascending = False, inplace = True)
			best = results.iloc[0]

			if taxon == "species":
				if best.ani < best.threshold:
					return f"{best.species}*({best.ani})"
				else:
					return f"{best.species}({best.ani})"

			elif taxon == "subspecies":
				if best.ani < best.threshold:
					return "No subspecies"
				else:
					return f"{best.subspecies}({best.ani})"

			elif taxon == "geneflow":
				if best.ani < best.threshold:
					# if the ANI value doesn't fall within the pseudo-gene flow 
					# unit ANI boundary, try the 2nd to 5th most similar genomes
					for candidate in itertools.islice(results.itertuples(), 1, 5):
						if candidate.ani >= candidate.threshold:
							return f"{candidate.psub}({candidate.ani})"
					return f"{best.psub}*({best.ani})"
				else:
					return f"{best.psub}({best.ani})"

			elif taxon == "typestrains":
				return f"{best.typestrain}({best.ani})"

			elif taxon == "cytotoxicity":
				return f"{best.cytotoxicity_strain}({best.ani})"

		else:
			if taxon == "species":
				return "(Species unknown)"
			elif taxon == "subspecies":
				return "No subspecies"
			elif taxon == "geneflow":
				return "(Pseudo-gene flow unit unknown)"
			elif taxon == "typestrains":
				return "(Type strain unknown)"
			elif taxon == "cytotoxicity":
				return "(Cytotoxicity strain unknown)"

	def check_fragmentation(self, sequences, fragment_length):
		length_total = 0
		length_in_fragment = 0
		for sequence in sequences:
			length_total += len(sequence)
			length_in_fragment += len(sequence) % fragment_length
		if length_in_fragment * 2 > length_total:
			warnings.warn("Genome is heavily fragmented, ANI results will not be reliable.")



import os
import subprocess

import pandas as pd
from pandas.errors import EmptyDataError

class Blast:
	"""
	Detect genes with blast

	run_blast
	purpose:
		blast a query against a reference and save results to a text file
	input:
		task = blast algorithm to use (e.g., tblastn)
		dbseqs = blast database fasta file
		fasta = blast query fasta file
		final_results_directory = BRiskTyper final results directory
		prefix = genome prefix for output files
		suffix = suffix for output files corresponding to BRiskTyper task (e.g., bt, virulence)
		pthresh = minimum percent identity threshold for blast hits
		qthresh = minimum percent query coverage for blast hits
		overlap = maximum proportion of overlap for overlapping blast hits to be considered separate (hits overlapping below this threshold will be considered separate hits)
		evalue = maximum blast e-value for blast hits

	output:
		path to blast output file of results

	parse_virulence
	purpose:
		assign query genome to virulence-associated biovars, using blast results from run_blast
	input:
		virfile = path to run_blast results file containing virulence gene hits
		pthresh = minimum percent identity threshold for blast hits
		qthresh = minimum percent query coverage for blast hits
	output:
		lists of virulence genes detected for each virulence-associated biovar

	parse_bt
	purpose:
		assign query genome to biovar Thuringiensis or not, using blast results from run_blast
	input:
		btfile = path to blast results file containing Bt toxin gene hits
		pthresh = minimum percent identity threshold for blast hits
		qthresh = minimum percent query coverage for blast hits
		overlap = maximum proportion of overlap for overlapping Bt toxin genes to be considered separate genes
			Bt toxin genes below this threshold will be considered separate, while those above it will be considered overlapping, and only the top hit will be reported
	output:
		a sorted list of Bt genes detected in a sequence

	parse_mlst
	purpose:
		assign query genome to sequence type (ST), using multi-locus sequence typing (MLST) and blast results from run_blast
	input:
		mlstfile = path to blast results file containing MLST gene hits
	output:
		list of top hits for each allele
	"""

	def __init__(self, task, dbseqs, fasta, final_results_directory, prefix, suffix, pthresh, qthresh, overlap, evalue):

		self.task = task
		self.dbseqs = dbseqs
		self.fasta = fasta
		self.final_results_directory = final_results_directory
		self.prefix = prefix
		self.suffix = suffix
		self.pthresh = pthresh
		self.qthresh = qthresh
		self.overlap = overlap
		self.evalue = evalue

	def run_blast(self, task, dbseqs, fasta, final_results_directory, prefix, suffix, evalue):
		# create output folder if it doesn't exist
		blast_results_dir = os.path.join(final_results_directory, suffix)
		os.makedirs(blast_results_dir, exist_ok=True)

		# build a BLAST database if it doesn't exist yet
		if not os.path.exists(dbseqs + ".nsq"):
			proc = subprocess.run(["makeblastdb", "-in", dbseqs, "-dbtype", "nucl"])
			proc.check_returncode()

		# run the BLAST task
		blast_results = os.path.join(blast_results_dir, "{}_{}.txt".format(prefix, suffix))
		proc = subprocess.run([
			task,
			"-query", fasta,
			"-db", dbseqs,
			"-out", blast_results,
			"-max_target_seqs", "1000000000",
			"-evalue", evalue,
			"-outfmt", "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen qcovs qcovhsp",
		])
		proc.check_returncode()

		# return path to results
		return blast_results

	def parse_virulence(self, virfile, pthresh, qthresh):

		try:

			blast_results_file = pd.read_csv(virfile, sep = "\s+", header = None)
			blast_results_file = blast_results_file.sort_values(by = [0], ascending = True) # genome as query and mlst db as db
			genes = blast_results_file.iloc[:,0] # genome as db and mlst db as query

			emetic = []
			anthracis = []
			nhe = []
			hbl = []
			cytK = []
			sph = []
			cap = []
			has = []
			bps = []
			max_cyt = 0
			genes = genes.unique()
			for gene in genes:
				gene_subset = blast_results_file[blast_results_file[0].str.match(gene)] # genome as query and mlst db as db
				gene_subset = gene_subset.sort_values(by = [11], ascending = False)
				max_gene = gene_subset.iloc[0,0]
				pid = float(gene_subset.iloc[0,2])
				qid = float(gene_subset.iloc[0,14])
				#qid = 100*(float(gene_subset.iloc[0,3])/float(gene_subset.iloc[0,12]))
				if pid >= pthresh and qid >= qthresh:
					if max_gene == "cesA" or max_gene == "cesB" or max_gene == "cesC" or max_gene == "cesD":
						emetic.append(max_gene)
					if max_gene == "cya" or max_gene == "lef" or max_gene == "pagA":
						anthracis.append(max_gene)
					if max_gene == "nheA" or max_gene == "nheB" or max_gene == "nheC":
						nhe.append(max_gene)
					if max_gene == "hblA" or max_gene == "hblB" or max_gene == "hblC" or max_gene == "hblD":
						hbl.append(max_gene)
					if max_gene == "cytK-1" or max_gene == "cytK-2":
						cyt_bits = gene_subset.iloc[0,11]
						if cyt_bits > max_cyt:
							max_cyt = cyt_bits
							cytK = []
							cytK.append(max_gene)
					if max_gene == "sph":
						sph.append(max_gene)
					if max_gene == "capA" or max_gene == "capB" or max_gene == "capC" or max_gene == "capD" or max_gene == "capE":
						cap.append(max_gene)
					if max_gene == "hasA" or max_gene == "hasB" or max_gene == "hasC":
						has.append(max_gene)
					if max_gene == "bpsA" or max_gene == "bpsB" or max_gene == "bpsC" or max_gene == "bpsD" or max_gene == "bpsE" or max_gene == "bpsF" or max_gene == "bpsG" or max_gene == "bpsH" or max_gene == "bpsX":
						bps.append(max_gene)

		except EmptyDataError:

			anthracis = []
			emetic = []
			nhe = []
			hbl = []
			cytK = []
			sph = []
			cap = []
			has = []
			bps = []

		return(anthracis, emetic, nhe, hbl, cytK, sph, cap, has, bps)



	def parse_bt(self, btfile, pthresh, qthresh, overlap):

		try:

			blast_results_file = pd.read_csv(btfile, sep = "\s+", header = None)
			blast_results_file = blast_results_file.sort_values(by = [0], ascending = True) # genome as query and mlst db as db
			genes = blast_results_file.iloc[:,0] # genome as db and mlst db as query

			rangedict = {}
			bitdict = {}
			genes = genes.unique()
			for gene in genes:
				gene_subset = blast_results_file[blast_results_file[0].str.match(gene)] # genome as query and mlst db as db
				gene_subset = gene_subset.sort_values(by = [11], ascending = False)
				max_gene = gene_subset.iloc[0,0]
				pid = float(gene_subset.iloc[0,2])
				qid = float(gene_subset.iloc[0,14])
				#qid = 100*(float(gene_subset.iloc[0,3])/float(gene_subset.iloc[0,12]))
				glen = float(gene_subset.iloc[0,3])
				gstart = int(gene_subset.iloc[0,8])
				gend = int(gene_subset.iloc[0,9])
				if gstart < gend:
					grange = list(range(gstart, gend + 1))
				elif gend < gstart:
					grange = list(range(gend, gstart + 1))
				if pid >= pthresh and qid >= qthresh:
					rangedict[max_gene] = grange
					bitdict[max_gene] = float(gene_subset.iloc[0,11])


			bt = []
			#comparisons = []
			for key, val in rangedict.items():
				overlap_dict = {}
				test_bits = bitdict[key]
				grange = val
				for key2, val2 in rangedict.items():
					if key != key2: #and key + "VS" + key2 not in comparisons and key2 + "VS" + key not in comparisons:
						#comparisons.append(key + "VS" + key2)
						#comparisons.append(key2 + "VS" + key)
						test_overlap = float(len(list(set(grange) & set(val2))))/float(len(val2))
						og_bits = bitdict[key2]
						if test_overlap > overlap: #and test_bits  > og_bits:
							if key not in overlap_dict.keys():
								overlap_dict[key] = test_bits
							overlap_dict[key2] = og_bits
				if len(overlap_dict.keys()) > 0:
					maxbits = 0
					max_gene = []
					for okey in overlap_dict.keys():
						oval = overlap_dict[okey]
						if oval > maxbits:
							maxbits = oval
							max_gene = [okey]
						elif oval == maxbits:
							max_gene.append(okey)
				else:
					max_gene = [key]
				max_gene = sorted(max_gene)
				if len(max_gene) > 1:
					max_gene = "/".join(max_gene)
				else:
					max_gene = max_gene[0]
				if max_gene not in bt:
					bt.append(max_gene)

		except EmptyDataError:
			bt = []
		return(sorted(bt))


	def parse_mlst(self, mlstfile):

		try:

			blast_results_file = pd.read_csv(mlstfile, sep = "\s+", header = None)
			blast_results_file = blast_results_file.sort_values(by = [0], ascending = True) # genome as query and mlst db as db
			genes = blast_results_file.iloc[:,0] # genome as db and mlst db as query

			mlst = []
			perfect_matches = 0

			genes = genes.str.split("_",expand=True)
			genes = genes.iloc[:,0]
			genes = genes.unique()
			for gene in genes:
				gene_subset = blast_results_file[blast_results_file[0].str.contains(gene)] # genome as query and mlst db as db
				gene_subset = gene_subset.sort_values(by = [11], ascending = False)
				max_bits = gene_subset.iloc[0,11]
				pid = float(gene_subset.iloc[0,2])
				qid = float(gene_subset.iloc[0,14])
				if pid == 100.0 and qid == 100.0:
					perfect_matches += 1
				max_genes = gene_subset[gene_subset[11]==max_bits]
				max_genes = max_genes.iloc[:,0]
				mlst.append([mg.split("_")[-1].strip() for mg in set(max_genes)])

		except EmptyDataError:
			mlst = []
			perfect_matches = 0

		return(mlst, perfect_matches)




	def parse_panC(self, panCfile):

		try:

			blast_results_file = pd.read_csv(panCfile, sep = "\s+", header = None)
			blast_results_file = blast_results_file.sort_values(by = [11], ascending = False)
			max_bits = blast_results_file.iloc[0,11]
			pid = float(blast_results_file.iloc[0,2])
			qid = float(blast_results_file.iloc[0,14])
			max_gene = blast_results_file.iloc[0,0]
			print(pid)
			print(qid)
			if pid < 99.0 or qid < 80.0:
					max_gene = max_gene + "*"


		except EmptyDataError:
			max_gene = "Unknown(missing allele)"

		return(max_gene)



#!/usr/bin/env python3

import argparse
import contextlib
import datetime
import importlib.resources
import logging
import os
import sys
import shutil
import tempfile
import urllib.request
import warnings
import xml.etree.ElementTree as etree

from Bio import SeqIO

from .ani import Ani
from .blast import Blast
from .mlst import Mlst
from .print_final_results import FinalResults

__author__ = "Laura M. Carroll <lmc297@cornell.edu>"
__version__ = "0.1.10"


@contextlib.contextmanager
def _forward_warnings():
	_showwarning = warnings.showwarning
	try:
		warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(f"Warning: {message}")
		yield
	finally:
		warnings.showwarning = _showwarning


def run_pipeline(args):
	"""
	run brisktyper
	"""

	# get path to brisktyper executable
	brisktyper_path = os.path.realpath(__file__)
	brisktyper_path = brisktyper_path.rpartition("/")[0].strip() + "/"

	# get input and output arguments and prefix to use for output files
	infile = args.input[0]
	prefix = infile.split("/")[-1].strip()
	prefix = ".".join(prefix.split(".")[0:-1])
	output = args.output[0]

	# add a / to output directory name, if one is not already supplied
	if not output.endswith("/"):
		output = output.strip() + "/"

	# if the output directory doesn't yet exist, make it
	if not os.path.isdir(output):
		os.mkdir(output)

	# if a brisktyper_final_results directory does not exist
	# in the output directory, make it
	if not os.path.isdir(output + "brisktyper_final_results"):
		os.mkdir(output + "brisktyper_final_results/")

	# make a logs directory for log files
	if not os.path.isdir(output + "brisktyper_final_results/logs"):
		os.mkdir(output + "brisktyper_final_results/logs")

	# save path to final results directory
	final_results_directory = output + "brisktyper_final_results/"

	# initialize log file
	logging.basicConfig(level = logging.DEBUG, filename = final_results_directory + "logs/" + prefix + ".log", filemode = "a+", format = "%(message)s")
	logging.getLogger().addHandler(logging.StreamHandler())


	# get arguments
	# evalue is stored as a string because it gets passed to blast command as a string
	ani_species = args.ani_species
	ani_subspecies = args.ani_subspecies
	ani_geneflow = args.ani_geneflow
	ani_typestrains = args.ani_typestrains
	ani_cytotoxicity = args.ani_cytotoxicity
	virulence = args.virulence
	bt = args.bt
	mlst = args.mlst
	panC = args.panC
	vdb = args.virulence_db
	vpthresh = int(args.virulence_identity)
	vqthresh = int(args.virulence_coverage)
	bpthresh = int(args.bt_identity)
	bqthresh = int(args.bt_coverage)
	overlap = float(args.bt_overlap)
	evalue = str(args.evalue)
	download_mlst_latest = args.download_mlst_latest

	# log to file
	now = datetime.datetime.now
	logging.info("Welcome to BRiskTyper!")
	logging.info("You are initializing this run at " + now().strftime("%Y-%m-%d %H:%M"))
	logging.info("You ran the following command: ")
	logging.info(" ".join([str(sa) for sa in sys.argv]))
	logging.info("Report bugs/concerns to Laura M. Carroll <laura.carroll@embl.de>")


	# perform species and/or supspecies assignment
	if ani_species == "True" or ani_subspecies == "True" or ani_geneflow == "True" or ani_typestrains == "True" or ani_cytotoxicity == "True":

		with _forward_warnings():

			if ani_species == "True":

				get_species = Ani(
					taxon = "species",
					fasta = infile,
					final_results_directory = final_results_directory,
					prefix = prefix)

				logging.info("Using PyFastANI to assign " + prefix + " to a species at " + now().strftime("%Y-%m-%d %H:%M"))
				final_species = get_species.run_fastani("species", infile, final_results_directory, prefix)
				logging.info("Finished species assignment of " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				final_species = "(Species assignment not performed)"

			if ani_subspecies == "True":

				get_subspecies = Ani(
					taxon = "subspecies",
					fasta = infile,
					final_results_directory = final_results_directory,
					prefix = prefix)

				logging.info("Using FastANI to assign " + prefix + " to a subspecies (if applicable) at " + now().strftime("%Y-%m-%d %H:%M"))
				final_subspecies = get_subspecies.run_fastani("subspecies", infile, final_results_directory, prefix)
				logging.info("Finished subspecies assignment of " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				final_subspecies = "(Subspecies assignment not performed)"

			if ani_geneflow == "True":

				get_geneflow = Ani(
					taxon = "geneflow",
					fasta = infile,
					final_results_directory = final_results_directory,
					prefix = prefix)

				logging.info("Using FastANI to assign " + prefix + " to a pseudo-gene flow unit at " + now().strftime("%Y-%m-%d %H:%M"))
				final_geneflow = get_geneflow.run_fastani("geneflow", infile, final_results_directory, prefix)
				logging.info("Finished pseudo-gene flow unit assignment of " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				final_geneflow = "(Pseudo-gene flow unit assignment not performed)"

			if ani_typestrains == "True":

				get_typestrains = Ani(
					taxon = "typestrains",
					fasta = infile,
					final_results_directory = final_results_directory,
					prefix = prefix)

				logging.info("Using FastANI to compare " + prefix + " to B. cereus s.l. species type strain genomes at " + now().strftime("%Y-%m-%d %H:%M"))
				final_typestrains = get_typestrains.run_fastani("typestrains", infile, final_results_directory, prefix)
				logging.info("Finished B. cereus s.l. species type strain comparison of " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				final_typestrains = "(Type strain-based taxonomic assignment not performed)"


			if ani_cytotoxicity == "True":

				get_cytotoxicity = Ani(
					taxon = "cytotoxicity",
					fasta = infile,
					final_results_directory = final_results_directory,
					prefix = prefix)

				logging.info("Using FastANI to compare " + prefix + " to cytotoxicity assay strains at " + now().strftime("%Y-%m-%d %H:%M"))
				final_cytotoxicity = get_cytotoxicity.run_fastani("cytotoxicity", infile, final_results_directory, prefix)
				logging.info("Finished B. cereus s.l. cytotoxicity assay comparison of " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				final_cytotoxicity = "(Cytotoxicity assay comparison not performed)"

	else:

		final_species = "(Species assignment not performed)"
		final_subspecies = "(Subspecies assignment not performed)"
		final_geneflow = "(Pseudo-gene flow unit assignment not performed)"
		final_typestrains = "(Type strain-based taxonomic assignment not performed)"
		final_cytotoxicity = "(Cytotoxicity assay comparison not performed)"



	# perform virulence-associated biovar assignment
	if virulence == "True":

		if vdb == "aa":
			vdb_name = "brisktyper_virulence_sequences.faa"
			vdb_task = "tblastn"

		elif vdb == "nuc":
			vdb_name = "brisktyper_virulence_sequences.ffn"
			vdb_task = "blastn"


		with importlib.resources.path("brisktyper.seq_virulence_db", vdb_name) as vdb_path:

			get_virulence = Blast(
				task = vdb_task,
				dbseqs = infile,
				fasta = vdb_path,
				final_results_directory = final_results_directory,
				prefix = prefix,
				suffix = "virulence",
				pthresh = vpthresh,
				qthresh = vqthresh,
				overlap = overlap,
				evalue = evalue)

			logging.info("Using " + vdb_task + " to identify potential virulence factors in " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			vir = get_virulence.run_blast(vdb_task, infile, vdb_path, final_results_directory, prefix, "virulence", evalue)
			anthracis, emetic, nhe, hbl, cytK, sph, cap, has, bps = get_virulence.parse_virulence(vir, vpthresh, vqthresh)

			logging.info("Finished virulence factor detection in " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))


	else:

		anthracis = "(Virulence factor detection not performed)"
		emetic = "(Virulence factor detection not performed)"
		nhe = "(Virulence factor detection not performed)"
		hbl = "(Virulence factor detection not performed)"
		cytK = "(Virulence factor detection not performed)"
		sph = "(Virulence factor detection not performed)"
		cap = "(Virulence factor detection not performed)"
		has = "(Virulence factor detection not performed)"
		bps = "(Virulence factor detection not performed)"



	# perform Thuringiensis biovar assignment
	if bt == "True":

		with importlib.resources.path("brisktyper.seq_bt_db", "brisktyper_bt_sequences.faa") as bt_path:

			get_bt = Blast(
				task = "tblastn",
				dbseqs = infile,
				fasta = bt_path,
				final_results_directory = final_results_directory,
				prefix = prefix,
				suffix = "bt",
				pthresh = bpthresh,
				qthresh = bqthresh,
				overlap = overlap,
				evalue = evalue)

			logging.info("Using tblastn to identify potential Bt genes in " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			bt_results = get_bt.run_blast("tblastn", infile, bt_path, final_results_directory, prefix, "bt", evalue)
			bt_final = get_bt.parse_bt(bt_results, bpthresh, bqthresh, overlap)

			logging.info("Finished Bt toxin gene detection for " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))


	else:

		bt_final = "(Bt toxin gene detection not performed)"




	# perform multi-locus sequence typing (MLST) using PubMLST's seven-gene scheme for Bacillus cereus
	if mlst == "True":

		with contextlib.ExitStack() as ctx:

			if download_mlst_latest == "True":

				logging.info("Downloading most recent PubMLST datbase at " + now().strftime("%Y-%m-%d %H:%M"))
				with urllib.request.urlopen("https://pubmlst.org/data/dbases.xml") as req:
					tree = etree.parse(req)
					parent = next(e for e in tree.iter("species") if e.text.strip() == "Bacillus cereus")
					urls = (e.text for e in parent.iter("url"))

				mlst_file = ctx.enter_context(tempfile.NamedTemporaryFile(suffix=".fas", mode="wb", buffering=0))
				bcereus_file = ctx.enter_context(tempfile.NamedTemporaryFile(suffix=".txt", mode="wb", buffering=0))
				mlst_path = mlst_file.name
				bcereus_path = bcereus_file.name

				for url in urls:
					if "alleles_fasta" in url:
						with urllib.request.urlopen(url) as req:
							shutil.copyfileobj(req, mlst_file)
					elif "profiles_csv" in url:
						with urllib.request.urlopen(url) as req:
							shutil.copyfileobj(req, bcereus_file)

				logging.info("Finished downloading most recent PubMLST datbase at " + now().strftime("%Y-%m-%d %H:%M"))

			else:

				db_time = importlib.resources.read_text("brisktyper.seq_mlst_db", "timestamp.txt").strip()
				logging.info("Using local PubMLST database (downloaded at {})".format(db_time))

				mlst_path = ctx.enter_context(importlib.resources.path("brisktyper.seq_mlst_db", "mlst.fas"))
				bcereus_path = ctx.enter_context(importlib.resources.path("brisktyper.seq_mlst_db", "bcereus.txt"))


			get_mlst = Blast(
				task = "blastn",
				dbseqs = infile,
				fasta = mlst_path,
				final_results_directory = final_results_directory,
				prefix = prefix,
				suffix = "mlst",
				pthresh = 0,
				qthresh = 0,
				overlap = overlap,
				evalue = evalue)

			logging.info("Using blastn to identify potential seven-gene MLST genes in " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			mlst_results = get_mlst.run_blast("blastn", infile, mlst_path, final_results_directory, prefix, "mlst", evalue)
			mlst_alleles, perfect_matches = get_mlst.parse_mlst(mlst_results)

			logging.info("Finished seven-gene MLST gene detection for " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			get_st = Mlst(
				alleles = mlst_alleles,
				profiles = bcereus_path,
				perfect_matches = perfect_matches,
				final_results_directory = final_results_directory,
				prefix = prefix)


			mlst_final = get_st.at2st(mlst_alleles, bcereus_path, perfect_matches, final_results_directory, prefix)


	else:

		mlst_final = "(Seven-gene MLST not performed)"



	# perform panC phylogenetic group assignment using the adjusted, eight-group panC group assignment scheme
	if panC == "True":

		with importlib.resources.path("brisktyper.seq_panC_db", "panC.fna") as panC_path:

			get_panC = Blast(
				task = "blastn",
				dbseqs = infile,
				fasta = panC_path,
				final_results_directory = final_results_directory,
				prefix = prefix,
				suffix = "panC",
				pthresh = 0,
				qthresh = 0,
				overlap = overlap,
				evalue = evalue)

			logging.info("Using blastn to identify panC in " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))

			panC_results = get_panC.run_blast("blastn", infile, panC_path, final_results_directory, prefix, "panC", evalue)
			panC_final = get_panC.parse_panC(panC_results)
			logging.info("Finished panC gene detection for " + prefix + " at " + now().strftime("%Y-%m-%d %H:%M"))


	else:

		panC_final = "(panC group assignment not performed)"



	# print results to a final results file
	get_final_results = FinalResults(
		final_results_directory = final_results_directory,
		infile = infile,
		prefix = prefix,
		species = final_species,
		subspecies = final_subspecies,
		geneflow = final_geneflow,
		typestrains = final_typestrains,
		cytotoxicity = final_cytotoxicity,
		anthracis = anthracis,
		emetic = emetic,
		nhe = nhe,
		hbl = hbl,
		cytK = cytK,
		sph = sph,
		cap = cap,
		has = has,
		bps = bps,
		bt_final = bt_final,
		mlst_final = mlst_final,
		panC_final = panC_final)

	get_final_results.print_final_results(final_results_directory, infile, prefix, final_species, final_subspecies, final_geneflow, final_typestrains, final_cytotoxicity, anthracis, emetic, nhe, hbl, cytK, sph, cap, has, bps, bt_final, mlst_final, panC_final)

	for blastdb_ext in ("nsq", "nin", "nhr", "ndb", "not", "ntf", "nto"):
		blastdb_file = "{}.{}".format(infile, blastdb_ext)
		if os.path.isfile(blastdb_file):
			os.remove(blastdb_file)

	logging.info("")
	logging.info("")
	logging.info("")
	logging.info("BRiskTyper finished at " + now().strftime("%Y-%m-%d %H:%M"))
	logging.info("Report bugs/concerns to Laura M. Carroll, laura.carroll@embl.de\n")
	logging.info("Have a nice day!")



def main():

	# BRiskTyper arguments

	parser = argparse.ArgumentParser(prog = "brisktyper", usage = "brisktyper -i </path/to/genome.fasta> -o </path/to/output/directory/> [other options]")

	parser.add_argument("-i", "--input", help = "Path to input genome in fasta format", nargs = 1, required = True)

	parser.add_argument("-o", "--output", help = "Path to desired output directory", nargs = 1, required = True)

	parser.add_argument("--ani_species", help = "Optional argument; True or False; assign genome to a species using FastANI; default = True", nargs = "?", default = "True")

	parser.add_argument("--ani_subspecies", help = "Optional argument; True or False; assign genome to a subspecies, if relevant, using FastANI; default = True", nargs = "?", default = "True")

	parser.add_argument("--ani_geneflow", help = "Optional argument; True or False; assign genome to a pseudo-gene flow unit using the method described by Carroll, et al. using FastANI; default = False", nargs = "?", default = "False")

	parser.add_argument("--ani_typestrains", help = "Optional argument; True or False; calculate ANI values between the query genome relative to all B. cereus s.l. species type strain genomes using FastANI, and report the closest species type strain/highest ANI value; default = True", nargs = "?", default = "True")

	parser.add_argument("--ani_cytotoxicity", help = "Optional argument; True or False; calculate ANI values between the query genome relative to all B. cereus s.l. strains used in a cytotoxicity assay, and report the closest strain/highest ANI value; default = True", nargs = "?", default = "True")

	parser.add_argument("--virulence", help = "Optional argument; True or False; perform virulence gene detection (required if one wants to assign genomes to biovars Anthracis or Emeticus); default = True", nargs = "?", default = "True")

	parser.add_argument("--bt", help = "Optional argument; True or False; perform Bt toxin gene detection for cry, cyt, and vip genes (required if one wants to assign genomes to biovar Thuringiensis); default = True", nargs = "?", default = "True")

	parser.add_argument("--mlst", help = "Optional argument; True or False; assign genome to a sequence type (ST) using the seven-gene multi-locus sequence typing (MLST) scheme available in PubMLST; default = True", nargs = "?", default = "True")

	parser.add_argument("--panC", help = "Optional argument; True or False; assign genome to a phylogenetic group (Group I-VIII) using an adjusted, eight-group panC group assignment scheme; default = True", nargs = "?", default = "True")

	parser.add_argument("--virulence_db", help = "Optional argument for use with --virulence True; aa or nuc; database to use for virulence factor detection: aa for the amino acid sequence database, or nuc for the nucleotide sequence database; option aa uses translated nucleotide blast and allows for the detection of more remote homologs, but is slower than nuc, which uses blastn; default = aa", nargs = "?", default = "aa")

	parser.add_argument("--virulence_identity", help = "Optional argument for use with --virulence True; integer from 0 to 100; minimum percent amino acid/nucleotide identity threshold for a virulence gene to be considered present, depending on choice of --virulence_db aa or nuc, respectively; default = 70", nargs = "?", default = 70)

	parser.add_argument("--virulence_coverage", help = "Optional argument for use with --virulence True; integer from 0 to 100; minimum percent coverage threshold for a virulence gene to be considered present; default = 80", nargs = "?", default = 80)

	parser.add_argument("--bt_identity", help = "Optional argument for use with --bt True; integer from 0 to 100; minimum percent amino acid identity threshold for a Bt toxin gene to be considered present; default = 50", nargs = "?", default = 50)

	parser.add_argument("--bt_coverage", help = "Optional argument for use with --bt True; integer from 0 to 100; minimum percent coverage threshold for a Bt toxin gene to be considered present; default = 70", nargs = "?", default = 70)

	parser.add_argument("--bt_overlap", help = "Optional argument for use with --bt True; float from 0 to 1; specify maximum proportion of overlap for overlapping Bt toxin genes to be considered separate genes; Bt toxin genes below this threshold will be considered separate, while those above it will be considered overlapping, and only the top hit will be reported; default=0.7", nargs = "?", default = 0.7)

	parser.add_argument("--evalue", help = "Optional argument for use with --virulence True and/or --bt True; float >= 0; maximum blast e-value for a hit to be saved; note that if both --virulence True and --bt True, this e-value threshold will be applied to both analyses; default = 1e-5", nargs = "?", default = 1e-5)

	parser.add_argument("--download_mlst_latest", help = "Optional argument for use with --mlst True; True or False; download the latest version of the seven-gene multi-locus sequence typing (MLST) scheme available in PubMLST; if this is False, BRiskTyper will search for the appropriate files in the seq_mlst_db directory; default = False", nargs = "?", default = "False")

	parser.add_argument("--version", action="version", version='%(prog)s {}'.format(__version__), help="Print version")

	args = parser.parse_args()

	run_pipeline(args)

if __name__ == "__main__":

	# run BRiskTyper

	main()
